/*
 *	UCW System Startup and Shutdown Utility
 *
 *	Copyright (c) 1998 Martin Mares <mj@ucw.cz>
 *	Copyright (c) 1998 Pavel Machek <pavel@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <ctype.h>
#include <dirent.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>

typedef unsigned char byte;

/*
 *	Arguments
 */

static char *arglist = "-bcdefikqrs:uvAEF:RS:019";

#ifdef DEBUG
static byte *conf_file = "daemons";
static byte *status_file = "status";
static byte *shell = "-";
#else
static byte *conf_file = "/etc/daemons";
static byte *status_file = "/var/run/daemons";
static byte *shell = "/bin/sh";
#endif

static int verbose;
static int reset_status;
static int force_action;
static int shell_async;

static void
help(void)
{
  fprintf(stderr, "Usage: sss [[<options>] [<action>] <subsystem>] ...\n\
Actions:\n\
\t-u\tBring selected subsystems up (default)\n\
\t-d (-0)\tBring selected subsystems down\n\
\t-i\tInit mode: Bring selected ones up, everything else down\n\
\t-b\tPre-reboot shutdown\n\
\t-k (-9)\tKill\n\
\t-c (-1)\tReload configuration
\t-r\tRestart\n\
\t-q\tQuery status\n\
\t-e\tIgnore given subtree\n\
Options:\n\
\t-v\tIncrease verbosity\n\
\t-F <f>\tRead configuration from specified file\n\
\t-s <f>\tSet shell to pipe commands to ('-' = dump to stdout)\n\
\t-S <f>\tUse specified status file\n\
\t-R\tReset status file\n\
\t-f\tForce action even if the status file says it's unnecessary\n\
\t-E\tIgnore everything listed in $except\n\
\t-A\tRun commands asynchronously\n\
");
  exit(1);
}

/*
 *	Prototypes
 */

static void reset_report(void);

/*
 *	Utility functions
 */

static void  __attribute__((noreturn))
die(char *msg, ...)
{
  va_list args;

  va_start(args, msg);
  fputs("sss: ", stderr);
  vfprintf(stderr, msg, args);
  fputc('\n', stderr);
  exit(0);
}

static void
warn(char *msg, ...)
{
  va_list args;

  va_start(args, msg);
  fputs("sss: warning: ", stderr);
  vfprintf(stderr, msg, args);
  fputc('\n', stderr);
  va_end(args);
}

static void
verb(char *msg, ...)
{
  va_list args;
  int verb = 2;

  va_start(args, msg);
  if (msg[0] == '#' && isdigit(msg[1]))
    {
      verb = msg[1] - '0';
      msg += 2;
    }
  if (verbose >= verb)
    {
      vfprintf(stderr, msg, args);
      fputc('\n', stderr);
    }
  va_end(args);
}

static void *
xmalloc(unsigned int size)
{
  void *x = malloc(size);

  if (!x)
    die("Unable to allocate %d bytes of memory.", size);
  return x;
}

#define CHUNK_SIZE 4080
#define CHUNK_ALIGN 8

static void *
lalloc(unsigned int size)
{
  static byte *pool;
  static unsigned int rem;
  void *z;

  size = (size + CHUNK_ALIGN - 1) & ~(CHUNK_ALIGN - 1);
  if (rem < size)
    {
      pool = xmalloc(CHUNK_SIZE);
      rem = CHUNK_SIZE;
    }
  z = pool;
  pool += size;
  rem -= size;
  return z;
}

static byte *
stralloc(byte *s)
{
  int l = strlen(s) + 1;
  byte *z = lalloc(l);
  memcpy(z, s, l);
  return z;
}

static byte *
base_name(byte *n)
{
  byte *z = n;

  while (*n)
    if (*n++ == '/')
      z = n;
  return z;
}

/*
 *	In-memory version of the configuration file.
 */

struct attr {
  struct attr *next, *extend;
  byte *name, *value;
};

struct node {
  struct node *next, *sub, *parent;
  struct attr *item;
  byte ignored;
  byte name[1];
};

static struct node root;
static int cf_read;

static struct node *
get_node(byte *w, int create)
{
  struct node *a = &root;
  struct node **b, *p;
  byte *z, *n;
  byte r[strlen(w) + 1];

  if (!w)
    return NULL;
  if (*w == '/')
    w++;
  strcpy(r, w);
  n = r;
  while (n)
    {
      if (z = strchr(n, '/'))
	*z++ = 0;
      if (!*n)
	break;
      b = &a->sub;
      p = a;
      while (a = *b)
	{
	  if (!strcmp(a->name, n))
	    goto next;
	  b = &a->next;
	}
      if (!create)
	return NULL;
      *b = a = lalloc(sizeof(struct node) + strlen(n));
      a->next = a->sub = NULL;
      a->parent = p;
      a->item = NULL;
      a->ignored = 0;
      strcpy(a->name, n);
    next:
      n = z;
    }
  return a;
}

static struct attr *
lookup_attr(struct node *n, byte *name)
{
  struct attr *a;

  for(a=n->item; a; a=a->next)
    if (!strcmp(a->name, name))
      break;
  return a;
}

static byte *
get_attr(struct node *n, byte *name)
{
  struct attr *a;

  if (!(a = lookup_attr(n, name)))
    return NULL;
  else
    return a->value;
}

static int
get_attr_bool(struct node *n, byte *name)
{
  byte *a;

  return (a = get_attr(n, name)) && strcmp(a, "0");
}

static struct attr *
set_attr(struct node *n, byte *name, byte *val)
{
  struct attr *a;

  if (!(a = lookup_attr(n, name)))
    {
      a = lalloc(sizeof(struct attr));
      a->next = n->item;
      n->item = a;
      a->name = name;
    }
  a->value = val;
  a->extend = NULL;
  return a;
}

static void
set_attr_num(struct node *n, byte *name, int val)
{
  struct attr *a;
  byte *z;

  switch (val)
    {
    case 0: z = "0"; break;
    case 1: z = "1"; break;
    default:
      die("set_attr_num: internal error");
    }
  if (!(a = lookup_attr(n, name)))
    set_attr(n, name, z);
  else
    a->value = z;
}

static byte *
do_get_node_path(struct node *n, byte *buf)
{
  byte *z = n->parent ? do_get_node_path(n->parent, buf) : buf;
  *z++ = '/';
  strcpy(z, n->name);
  return z + strlen(n->name);
}

static byte *
get_node_path(struct node *n)
{
  static byte *path_buffer;

  if (n == &root)
    return "/";
  if (!path_buffer)
    path_buffer = lalloc(256);
  *do_get_node_path(n, path_buffer) = 0;
  return path_buffer + 2;
}

/*
 *	Parsing of configuration file
 */

static int lino;

static void
cferr(char *msg)
{
  die("%s line %d: %s", conf_file, lino, msg);
}

static void
read_conf_file(void)
{
  FILE *f;
  char line[256];
  byte *z, *w;
  struct attr *attr = NULL;
  struct node *node = &root;

  if (cf_read)
    return;
  cf_read++;
  verb("Reading config file %s", conf_file);
  if (!(f = fopen(conf_file, "r")))
    die("Unable to open `%s': %m", conf_file);
  while (fgets(line, sizeof(line)-1, f))
    {
      lino++;
      if (strlen(line) >= sizeof(line) - 1)
	cferr("Line too long");
      if (z = strchr(line, '\n'))
	*z = 0;
      if (!line[0] || line[0] == '#')
	continue;
      if (!isspace(line[0]))
	{
	  for(z=line; *z && !isspace(*z); z++);
	  while (isspace(*z))
	    *z++ = 0;
	  node = get_node(line, 1);
	  if (node->item)
	    cferr("Already defined");
	  attr = set_attr(node, "args", stralloc(z));
	}
      else
	{
	  struct attr *a;
	  for(z=line; isspace(*z); z++);
	  if (!*z)
	    continue;
	  w = z;
	  while (*z && !isspace(*z))
	    {
	      if (*z == ':')
		{
		  *z++ = 0;
		  while (isspace(*z))
		    *z++ = 0;
		  attr = set_attr(node, stralloc(w), stralloc(z));
		  goto done;
		}
	      z++;
	    }
	  if (!attr)
	    cferr("Misplaced continuation");
	  a = lalloc(sizeof(struct attr));
	  attr->extend = a;
	  a->extend = a->next = NULL;
	  a->name = NULL;
	  a->value = stralloc(w);
	  attr = a;
	}
    done:
    }

  fclose(f);
}

/*
 *	The Status File
 */

static int sfd;
static int status_changed;
static int status_unlink;

static void
status_load(int ro)
{
  FILE *f;
  int fd, lino;
  byte buf[256];

  if (reset_status)
    return;
  if (ro)
    {
      verb("Reading status file %s read-only", status_file);
      sfd = open(status_file, O_RDONLY);
      if (sfd < 0)
	die("Unable to open status file (%s): %m", status_file);
      status_changed = -1;
    }
  else
    {
      verb("Reading status file %s", status_file);
      sfd = open(status_file, O_RDWR, 0666);
      if (sfd < 0)
	{
	  warn("status file (%s) not available (%m), assuming -R", status_file);
	  reset_status = 1;
	  return;
	}
      if (flock(sfd, LOCK_EX) < 0)
	die("Unable to lock status file: %m");
      if (status_unlink)
	{
	  verb("Unlinking status file");
	  if (unlink(status_file) < 0)
	    warn("cannot unlink status file: %m");
	  status_changed = -1;
	}
    }
  if (status_unlink)
    fd = sfd;
  else if ((fd = dup(sfd)) < 0)
    die("dup: %m");
  if (!(f = fdopen(fd, "r")))
    die("fdopen: %m");
  lino = 0;
  while (fgets(buf, sizeof(buf)-1, f))
    {
      byte *z;
      struct node *n;

      lino++;
      if (!(z = strchr(buf, '\n')))
	die("Status file corrupted at line %d", lino);
      *z = 0;
      verb("#4 ... %s", buf);
      if (!(n = get_node(buf, 0)) || !n->item)
	{
	  warn("%s no longer exists", buf);
	  continue;
	}
      set_attr_num(n, " up", 1);
    }
  fclose(f);
}

static int
node_is_up(struct node *n)
{
  return get_attr_bool(n, " up");
}

static void
stat_save_item(FILE *f, struct node *n, byte *buf, byte *p)
{
  strcpy(p, n->name);
  if (node_is_up(n))
    {
      fputs(buf, f);
      fputc('\n', f);
    }
  while (*p)
    p++;
  *p++ = '/';
  for(n=n->sub; n; n=n->next)
    stat_save_item(f, n, buf, p);
}

static void
status_save(void)
{
  FILE *f;
  struct node *n;
  byte buf[256];

  if (reset_status)
    {
      verb("Resetting status file");
      sfd = open(status_file, O_RDWR | O_CREAT, 0666);
      if (sfd < 0)
	die("Unable to create status file: %m");
    }
  if (status_changed <= 0)
    return;
  verb("Saving status file %s", status_file);
  if (lseek(sfd, 0, SEEK_SET) < 0 ||
      ftruncate(sfd, 0) < 0)
    die("Status file: %m");
  f = fdopen(sfd, "w");
  stat_save_item(f, &root, buf, buf);
  fclose(f);
}

/*
 *	Shell Interface
 */

static FILE *shell_in, *shell_out;
static pid_t shell_pid;

static void
shell_open(void)
{
  int sk[2];

  verb("#3Initializing shell %s", shell);
  if (!strcmp(shell, "-"))
    {
      shell_out = stdout;
      return;
    }
  if (socketpair(AF_UNIX, SOCK_STREAM, 0, sk) < 0)
    die("socketpair: %m");
  shell_pid = fork();
  if (shell_pid < 0)
    die("fork: %m");
  else if (!shell_pid)
    {
      close(0);
      dup(sk[0]);
      close(sk[1]);
      if (sfd > 0)
	close(sfd);
      setenv("PATH", "/sbin:/bin:/usr/sbin:/usr/bin", 1);
      execl(shell, shell, NULL);
      die("exec(%s): %m", shell);
    }
  else
    {
      close(sk[0]);
      if ((sk[0] = dup(sk[1])) < 0)
	die("dup: %m");
      if (!(shell_in = fdopen(sk[0], "r")) ||
	  !(shell_out = fdopen(sk[1], "w")))
	die("fdopen: %m");
    }
}

static void
shell_close(void)
{
  if (shell_in)
    {
      int st;
      verb("#3Waiting for shell exit");
      fclose(shell_out);
      fclose(shell_in);
      if (waitpid(shell_pid, &st, 0) != shell_pid)
	die("wait: %m");
      if (WIFSIGNALED(st))
	warn("%s died of signal %d", shell, WTERMSIG(st));
      else if (WIFEXITED(st))
	{
	  if (WEXITSTATUS(st))
	    warn("%s failed with exit code %d", shell, WEXITSTATUS(st));
	}
      else
	warn("%s died of unknown disease, status=0x%x", shell, st);
      shell_in = NULL;
    }
  shell_out = NULL;
}

static void
shell_send(char *fmt, ...)
{
  va_list args;

  va_start(args, fmt);
  if (!shell_out)
    shell_open();
  if (verbose >= 4)
    {
      fputs(">>> ", stdout);
      vfprintf(stdout, fmt, args);
      fputc('\n', stdout);
    }
  vfprintf(shell_out, fmt, args);
  fputc('\n', shell_out);
  va_end(args);
}

static void
shell_send_attr(struct attr *a)
{
  for(;a;a=a->extend)
    {
      byte *v = a->value;
      if (*v != '@')
	shell_send("%s", v);
      else switch (v[1])
	{
	case '>':
	  if (v[2])
	    shell_send("echo -e \"%s\"", v+2);
	  break;
        default:
	  warn("Unknown internal command `%s'", v);
	}
    }
}

static int
shell_flush(byte *msg)
{
  byte buf[16];
  int rc;

  if (!shell_in || shell_async)
    {
      if (shell_out)
	fflush(shell_out);
      return 1;
    }
  shell_send("echo >&0 @$?");
  fflush(shell_out);
  if (!fgets(buf, sizeof(buf)-1, shell_in) ||
      buf[0] != '@' ||
      sscanf(buf+1, "%d", &rc) != 1)
    {
      fprintf(stderr, "%s: shell died while waiting for completion\n", msg);
      shell_close();
      return 0;
    }
  if (!rc)
    return 1;
  fprintf(stderr, "%s: failed with rc=%d\n", msg, rc);
  reset_report();
  return 0;
}

/*
 *	PID Cache
 */

struct pid_entry {
  struct pid_entry *next;
  int pid;
  byte name[1];
};

static struct pid_entry *first_pid_entry;

static void
load_pid_cache(void)
{
  DIR *d;
  struct dirent *de;
  byte name[32];
  byte buf[1024];
  int fd, l;
  byte *k;
  struct pid_entry *p;

  verb("#3Loading PID cache");
  if (!(d = opendir("/proc")))
    return;
  while (de = readdir(d))
    {
      byte *k = de->d_name;
      if (strlen(k) > 10)
	continue;
      while (isdigit(*k))
	k++;
      if (*k)
	continue;
      sprintf(name, "/proc/%s/cmdline", de->d_name);
      if ((fd = open(name, O_RDONLY)) < 0)
	continue;
      if ((l = read(fd, buf, sizeof(buf) - 1)) > 0)
	{
	  buf[l] = 0;
	  if (k = strchr(buf, ' '))
	    *k = 0;
	  k = base_name(buf);
	  while (*k == '-')
	    k++;
	  if (*k)
	    {
	      p = lalloc(sizeof(struct pid_entry) + strlen(k));
	      p->next = first_pid_entry;
	      first_pid_entry = p;
	      p->pid = strtol(de->d_name, NULL, 10);
	      strcpy(p->name, k);
	      verb("#5PID(%s)=%d", p->name, p->pid);
	    }
	}
      close(fd);
    }
  closedir(d);
}

static struct pid_entry *
find_pid_entry(byte *name, struct pid_entry *p)
{
  if (p)
    p = p->next;
  else
    {
      if (!first_pid_entry)
	load_pid_cache();
      p = first_pid_entry;
    }
  while (p && strcmp(p->name, name))
    p = p->next;
  return p;
}

static int
find_pid(byte *name)
{
  struct pid_entry *p = find_pid_entry(name, NULL);
  return p ? p->pid : 0;
}

static void
kill_all(byte *name, int sig)
{
  struct pid_entry *p = NULL;
  int found = 0;

  name = base_name(name);
  while (p = find_pid_entry(name, p))
    if (p->pid)
      {
	verb("#2killing %d by %d\n", p->pid, sig);
	kill(p->pid, sig);
	p->pid = 0;
	found = 1;
      }
  if (!found)
    warn("%s: not running", name);
}

/*
 *	Actions
 */

typedef enum action {
  A_NULL,
  A_UP,
  A_DOWN,
  A_INIT,
  A_KILL,
  A_RELOAD,
  A_RESTART,
  A_QUERY,
  A_REBOOT,
  A_INIT_UP,
  A_INIT_DOWN
} action;

#define M_NULL (1 << A_NULL)
#define M_UP (1 << A_UP)
#define M_DOWN (1 << A_DOWN)
#define M_INIT (1 << A_INIT)
#define M_KILL (1 << A_KILL)
#define M_RELOAD (1 << A_RELOAD)
#define M_RESTART (1 << A_RESTART)
#define M_QUERY (1 << A_QUERY)
#define M_REBOOT (1 << A_REBOOT)
#define M_INIT_UP (1 << A_INIT_UP)
#define M_INIT_DOWN (1 << A_INIT_DOWN)

#define ACT_IN(a, set) ((1 << (a)) & (set))

static char *act_msgs[] = {
  NULL,
  "Starting",
  "Stopping",
  "Initing",
  "Killing",
  "Reloading",
  "Restarting",
  "Querying",
  "Shutting down",
  "Starting",
  "Stopping"
};

struct act {
  struct act *next;
  action a;
  struct node *node;
};

static struct act *first_act = NULL;
static struct act **last_act = &first_act;
static unsigned int used_actions;

static void do_action(action, struct node *, int);

static void
queue_action(action act, struct node *n, byte *name)
{
  struct act *a;

  verb("#3Queued: %s %s", act_msgs[act], name);
  if (!act)
    {
      n->ignored = 1;
      return;
    }
  a = lalloc(sizeof(struct act));
  *last_act = a;
  last_act = &a->next;
  a->a = act;
  a->node = n;
  return;
}

static int
new_act_acronym(struct node *n, byte *name, action act)
{
  int r = 0;

  if (!strcmp(n->name, name))
    {
      queue_action(act, n, name);
      r++;
    }
  for(n=n->sub; n; n=n->next)
    r += new_act_acronym(n, name, act);
  return r;
}

static int
do_new_action(byte *name, action act)
{
  struct node *n;

  if (n = get_node(name, 0))
    {
      queue_action(act, n, name);
      return 1;
    }
  if (!strchr(name, '/'))
    return new_act_acronym(&root, name, act);
  return 0;
}

static void
new_action(byte *name, action act)
{
  used_actions |= (1 << act);
  if (!do_new_action(name, act))
    die("Unknown subsystem `%s'.", name);
}

static action last_action_rep = A_NULL;

static void
report_end(void)
{
  if (last_action_rep)
    shell_send("echo");
  last_action_rep = A_NULL;
}

static void
report(action a, struct node *n)
{
  byte *name = n->name;
  int chg = (last_action_rep != a);
  int single = !!get_attr(n, "newline");

  if (get_attr(n, "silent") || ACT_IN(a, M_QUERY | M_INIT))
    a = A_NULL;
  if (a)
    shell_send("echo -en \"%s%s%s%s%s\"",
	       (chg && last_action_rep) ? "\\n" : "",
	       chg ? act_msgs[a] : "",
	       chg ? ": " : "",
	       name,
	       single ? "\\n" : " ");
  else
    report_end();
  last_action_rep = single ? A_NULL : a;
}

static void
reset_report(void)
{
  last_action_rep = A_NULL;
}

static void
do_subtree(action a, struct node *n)
{
  for(n=n->sub; n; n=n->next)
    do_action(a, n, 0);
}

static void
do__subtree_rev(action a, struct node *n)
{
  if (n)
    {
      do__subtree_rev(a, n->next);
      do_action(a, n, 0);
    }
}

static inline void
do_subtree_rev(action a, struct node *n)
{
  do__subtree_rev(a, n->sub);
}

#define QS_UP 1
#define QS_DOWN 2

static int
do_query_scan(struct node *n)
{
  int z = 0;
  struct node *m;

  if (n->item)
    z |= (node_is_up(n) ? QS_UP : QS_DOWN);
  for(m=n->sub; m; m=m->next)
    z |= do_query_scan(m);
  return z;
}

static void
do_query(struct node *n)
{
  int z = do_query_scan(n);
  struct node *m;

  if (z == (QS_UP | QS_DOWN) || verbose)
    {
      for(m=n->sub; m; m=m->next)
	do_query(m);
      if (n->item)
	printf("%s: %s\n", get_node_path(n), node_is_up(n) ? "up" : "down");
    }
  else
    printf("%s: %s\n", get_node_path(n), (z == QS_UP) ? "up" : "down");
}

static void
init_action(action a, struct node *n, struct attr *x)
{
  struct attr *y;

  if (x && !strncmp(x->value, "@>", 2))
    report_end();
  else
    report(a, n);
  if (y = lookup_attr(n, "common"))
    shell_send_attr(y);
}

static void
do_action(action a, struct node *n, int explicit)
{
  struct node *m;

  if (n->ignored)
    {
      verb("Ignoring %s", get_node_path(n));
      return;
    }
  if (n->item)
    {
      byte *name = n->name;
      byte *args = get_attr(n, "args") ? : (byte *) "";
      byte *path = get_attr(n, "path") ? : name;
      struct attr *x;
      int up = node_is_up(n);
      int nup = up;
      int force = force_action || get_attr_bool(n, "force");
      int is_daemon = !get_attr(n, "up") && !get_attr(n, "down");

      verb("#2%s: %s", act_msgs[a], get_node_path(n));
      switch (a)			/* Real */
	{
	case A_INIT_UP:
	  if (up || !get_attr_bool(n, " init"))
	    break;
	  /* Fall-thru */
	case A_UP:
	  if (x = lookup_attr(n, "up"))
	    {
	      init_action(a, n, x);
	      shell_send_attr(x);
	    }
	  else if (is_daemon && (!up || !find_pid(base_name(path))))
	    {
	      if (x = lookup_attr(n, "start"))
		{
		  init_action(a, n, x);
		  shell_send_attr(x);
		}
	      else
		{
		  init_action(a, n, NULL);
		  shell_send("%s %s", path, args);
		}
	    }
	  if (shell_flush(name))
	    nup = 1;
	  do_subtree(a, n);
	  break;
	case A_INIT_DOWN:
	  if (!up || get_attr_bool(n, " init"))
	    break;
	  /* Fall-thru */
	case A_DOWN:
	  do_subtree_rev(a, n);
	  if (up || force)
	    {
	      if (!is_daemon)
		{
		  if ((x = lookup_attr(n, "down")) ||
		      (x = lookup_attr(n, "shutdown")))
		    {
		      init_action(a, n, x);
		      shell_send_attr(x);
		      if (shell_flush(name))
			nup = 0;
		    }
		}
	      else if ((x = lookup_attr(n, "stop")) ||
		       (x = lookup_attr(n, "shutdown")))
		{
		  init_action(a, n, x);
		  shell_send_attr(x);
		  if (shell_flush(name))
		    nup = 0;
		}
	      else
		{
		  init_action(a, n, NULL);
		  kill_all(path, SIGTERM);
		  nup = 0;
		}
	    }
	  else if (explicit)
	    warn("%s: not running", path);
	  break;
	case A_KILL:
	  do_subtree_rev(a, n);
	  if (up || force)
	    {
	      if (x = lookup_attr(n, "kill"))
		{
		  init_action(a, n, x);
		  shell_send_attr(x);
		  if (shell_flush(name))
		    nup = 0;
		}
	      else if (is_daemon)
		{
		  init_action(a, n, NULL);
		  kill_all(path, SIGKILL);
		  nup = 0;
		}
	    }
	  else if (explicit)
	    warn("%s: not running", path);
	  break;
	case A_RELOAD:
	  if (up || force)
	    {
	      if (x = lookup_attr(n, "reload"))
		{
		  init_action(a, n, x);
		  shell_send_attr(x);
		  shell_flush(name);
		}
	      else if (is_daemon)
		{
		  init_action(a, n, NULL);
		  kill_all(path, SIGHUP);
		}
	    }
	  else if (explicit)
	    warn("%s: not running", path);
	  do_subtree(a, n);
	  break;
	case A_RESTART:
	  if (!up && !force)
	    {
	      if (explicit)		/* FIXME: Do we really want this? */
		{
		  init_action(a, n, NULL);
		  do_action(A_UP, n, explicit);
		}
	    }
	  else if (x = lookup_attr(n, "restart"))
	    {
	      init_action(a, n, x);
	      shell_send_attr(x);
	      shell_flush(name);
	      do_subtree(a, n);
	    }
	  else
	    {
	      do_action(A_DOWN, n, explicit);
	      do_action(A_UP, n, explicit);
	    }
	  break;
	case A_REBOOT:
	  do_subtree_rev(a, n);
	  if (x = lookup_attr(n, "shutdown"))
	    {
	      init_action(a, n, x);
	      shell_send_attr(x);
	      shell_flush(name);
	    }
	  break;
	case A_INIT:
	  set_attr_num(n, " init", 1);
	  do_subtree(a, n);
	  break;
	case A_NULL:
	case A_QUERY:
	  die("This cannot happen, ask a wizzard!");
	}
      if (up != nup)
	{
	  set_attr_num(n, " up", nup);
	  if (status_changed >= 0)
	    status_changed = 1;
	}
    }
  else
    switch (a)				/* Only a subtree */
      {
      case A_UP:
      case A_INIT:
      case A_RELOAD:
      case A_RESTART:
      case A_INIT_UP:
	do_subtree(a, n);
	break;
      case A_DOWN:
      case A_KILL:
      case A_REBOOT:
      case A_INIT_DOWN:
	do_subtree_rev(a, n);
	break;
      case A_NULL:
      case A_QUERY:
	die("This cannot happen, ask a wizzard!");
      }
}

static void
do_actions(void)
{
  struct act *a;

  for(a=first_act; a; a=a->next)
    if (a->a == A_QUERY)
      do_query(a->node);
    else
      do_action(a->a, a->node, 1);
}

static void
do_init_actions(void)
{
  do_action(A_INIT_DOWN, &root, 1);
  do_action(A_INIT_UP, &root, 1);
}

/*
 *	The Central Cavern
 */

int
main(int argc, char **argv)
{
  int o;
  int stop = 0;
  action act = A_UP;

  while ((o = getopt(argc, argv, arglist)) >= 0)
    switch (o)
      {
      case 'u':
	act = A_UP; break;
      case 'd':
      case '0':
	act = A_DOWN; break;
      case 'i':
	act = A_INIT; break;
      case '9':
      case 'k':
	act = A_KILL; break;
      case 'c':
      case '1':
	act = A_RELOAD; break;
      case 'r':
	act = A_RESTART; break;
      case 'q':
	act = A_QUERY; break;
      case 'b':
	status_unlink = 1;
	act = A_REBOOT; break;
      case 'v':
	verbose++; break;
      case 'F':
	if (cf_read)
	  die("You must set config file name before any actions.");
	conf_file = optarg; break;
      case 'S':
	status_file = optarg; break;
      case 'R':
	reset_status = 1; break;
      case 'f':
	force_action = 1; break;
      case 'e':
	act = A_NULL; break;
      case 'E':
	{
	  byte *y, *x = getenv("except");
	  if (x)
	    {
	      byte buf[strlen(x)+1];
	      strcpy(buf, x);
	      y = buf;
	      read_conf_file();
	      while (x = y)
		{
		  if (y = strchr(x, ':'))
		    *y++ = 0;
		  if (*x)
		    do_new_action(x, A_NULL);
		}
	    }
	  break;
	}
      case 's':
	shell = optarg; break;
      case 'A':
	shell_async = 1; break;
      case 1:
	read_conf_file();
	new_action(optarg, act);
	break;
      default:
	help();
      }

  *last_act = NULL;
  if (!first_act)
    {
      if (ACT_IN(act, M_QUERY | M_REBOOT))
	{
	  read_conf_file();
	  new_action("/", act);
	  *last_act = NULL;
	}
      else
	help();
    }

  read_conf_file();
  status_load(!(used_actions & ~(M_QUERY | M_NULL)));
  do_actions();
  if (used_actions & (1 << A_INIT))
    do_init_actions();
  report_end();
  status_save();
  shell_close();
  return 0;
}
